/*
    Copyright 2017-2018 Intel Corporation.

    The source code, information and material ("Material") contained herein is owned by
    Intel Corporation or its suppliers or licensors, and title to such Material remains
    with Intel Corporation or its suppliers or licensors. The Material contains
    proprietary information of Intel or its suppliers and licensors. The Material is
    protected by worldwide copyright laws and treaty provisions. No part of the Material
    may be used, copied, reproduced, modified, published, uploaded, posted, transmitted,
    distributed or disclosed in any way without Intel's prior express written permission.
    No license under any patent, copyright or other intellectual property rights in the
    Material is granted to or conferred upon you, either expressly, by implication,
    inducement, estoppel or otherwise. Any license under such intellectual property
    rights must be express and approved by Intel in writing.

    Unless otherwise agreed by Intel in writing, you may not remove or alter this notice
    or any other notice embedded in Materials by Intel or Intel's suppliers or licensors
    in any way.
*/

#ifndef __PSTL_parallel_impl_H
#define __PSTL_parallel_impl_H

#include <atomic>
// This header defines the minimum set of parallel routines required to support Parallel STL,
// implemented on top of Intel(R) Threading Building Blocks (Intel(R) TBB) library

namespace pstl {
namespace internal {

//------------------------------------------------------------------------
// parallel_find
//-----------------------------------------------------------------------
/** Return extremum value returned by brick f[i,j) for subranges [i,j) of [first,last)
Each f[i,j) must return a value in [i,j). */
template<class _Index, class _Brick, class _Compare>
_Index parallel_find(_Index __first, _Index __last, _Brick __f, _Compare __comp, bool __b_first) {
    typedef typename std::iterator_traits<_Index>::difference_type _DifferenceType;
    const _DifferenceType __n = __last - __first;
    _DifferenceType __initial_dist = __b_first ? __n : -1;
    std::atomic<_DifferenceType> __extremum(__initial_dist);
    // TODO: find out what is better here: parallel_for or parallel_reduce
    par_backend::parallel_for(__first, __last, [__comp, __f, __first, &__extremum](_Index __i, _Index __j) {
        // See "Reducing Contention Through Priority Updates", PPoPP '13, for discussion of
        // why using a shared variable scales fairly well in this situation.
        if (__comp(__i - __first, __extremum)) {
            _Index __res = __f(__i, __j);
            // If not '__last' returned then we found what we want so put this to extremum
            if (__res != __j) {
                const _DifferenceType __k = __res - __first;
                for (_DifferenceType __old = __extremum; __comp(__k, __old); __old = __extremum) {
                    __extremum.compare_exchange_weak(__old, __k);
                }
            }
        }
    });
    return __extremum != __initial_dist ? __first + __extremum : __last;
}

//------------------------------------------------------------------------
// parallel_or
//------------------------------------------------------------------------
//! Return true if brick f[i,j) returns true for some subrange [i,j) of [first,last)
template<class _Index, class _Brick>
bool parallel_or(_Index __first, _Index __last, _Brick __f) {
    std::atomic<bool> __found(false);
    par_backend::parallel_for(__first, __last, [__f, &__found](_Index __i, _Index __j) {
        if (!__found.load(std::memory_order_relaxed) && __f(__i, __j)) {
            __found.store(true, std::memory_order_relaxed);
            par_backend::cancel_execution();
        }
    });
    return __found;
}

} // namespace internal
} // namespace pstl

#endif /* __PSTL_parallel_impl_H */
